import numpy as np
import scipy.linalg
def invert2( h, x0, q, y, r):
    pint = np.diag( q*q)
    rint = np.diag( r*r)
    k = np.dot(h, np.dot(pint, h.transpose())) + rint
    resid = y -np.dot( h, x0)
    x = np.dot(pint, np.dot( h.transpose(), np.dot( syminv(k), resid)))
    return x,pint
def invert( h, x0, q, y, r):
    """ calculates error reduction, uses two formulae depending on under or over determined """
    if h.shape[1] > h.shape[0]: # under determined
        print 'error_reduction, underdetermined'
        # expression is p = p0 -reduc where reduc = p0H^T (HP0H^T +R)^-1 H P0
        for i in xrange(q.size): h[:, i] *= q[i]
        k = np.dot( h,  h.transpose())
        for i in xrange(r.size): k[i,i] += r[i]**2
        for i in xrange(q.size): h[:, i] *= q[i]
        k = np.dot(h.transpose(), np.dot(syminv( k),  h))
        k  = np.diag( q* q) -k

        # now restore h
        for i in xrange(q.size): h[:, i] /= q[i]*q[i]
    else: # over-determined
        print ',error_reduction, overdetermined'
        # expression is p^-1 = P0^-1 + H^T R^-1 H
        for i in range( r.size): h[i,:] /= r[i]

        k = np.dot(h.transpose(), h)
        for i in range(q.size): k[i,i] += 1./(q[i]*q[i])
        k = syminv( k)
    # now we have the posterior covariance we can use fairly simple expressions for the posterior estimate
    x = np.dot( k, x0/q**2 + np.dot( y /r**2, h))

    return  x,k

from scipy.linalg import cho_factor, cho_solve
def syminv(a): return cho_solve( cho_factor( a), np.identity( a.shape[0]))

def read_kettle( filename):
    # reads the kettle budget with attached
    data = np.genfromtxt( filename, skip_header=1, dtype=None, delimiter='\t')
    vals = {}
    uncs = {}
    for i,name in enumerate(data['f0']):
        vals[name] = data['f1'][i]
        uncs[name] = np.abs(data['f3']-data['f2'])[i]/4.
    return vals,uncs


def normal_ineq(g1,g2, samplesize=1000000):
    """ uses montecarlo test for probability that random number from normal distribution g1 > that from g2"""
    v1 = np.random.normal( g1[0], g1[1], samplesize)
    v2 = np.random.normal(g2[0], g2[1], samplesize)
    return float((v1 > v2).sum())/float(samplesize)
    
kettleval, kettleunc = read_kettle('Kettle2002.txt')


# inverse problem for cos
# state vector contains fluxes to atm and inverse residence times for various removal processes
# observations include fluxes for removal processes and the requirement of a balanced budget
unknames = ['fococs', 'focdms', 'foccs2', 'fantocs', 'fantcs2', 'fantdms', 'ffire', 'focphot', 'kappaoh', 'kappacanopy', 'kappasoil']
obsnames = ['modernsinkoh', 'modernsinkcanopy', 'modernsinksoil', 'budget']

fococs =  [39.0,  39.0,   39.0]
focdms =  [81.0,  81.0,   81.0]
foccs2 =  [156.0, 156.0,  156.0]
fantocs = [0.0,   0.0,    64.0]
fantcs2 = [0.0,   0.0,    116.0]
fantdms = [0.0,   0.0,    0.5]    
ffire =   [136.0, 136.0,  136.0] 
focphot = [600.0, 600.0,  600.0]
conc = [380.0, 330.0, 476.0]
modernsinkoh = 101.0
modernsinkcanopy = 738.0
modernsinksoil = 355.0 
budget = 0.0


kappaoh = np.repeat( modernsinkoh/conc[-1] ,3)
kappacanopy = np.repeat( modernsinkcanopy/conc[-1] ,3)
kappasoil = np.repeat( modernsinksoil/conc[-1] ,3)

# now populate the prior and unc 
x0mod = np.zeros( len( unknames))
pmod = np.zeros( len( unknames))
for i,v in enumerate( unknames):
    x0mod[ i] = globals()[ v][-1]
    try: pmod[i] = kettleunc[v]
    except: pmod[i] = 0.
pmod[-3:] = 1. # infinite unc for kappas which we'll constrain from the observed fluxes



ymod = np.zeros( len( obsnames))
for i,v in enumerate( obsnames): ymod[ i] = globals()[ v]
rmod = 0.25 * ymod
rmod[-1] = 10. # tight balanced budget


hmod = np.zeros([ len( obsnames), len(unknames)]) # jacobian
# now fill bottom row for budget
hmod[-1,:-3] = 1. # jacobian element for fluxes
hmod[-1,-3:] = -conc[-1]
# now fill elements for th three fluxes
hmod[[0,1,2],[-3,-2,-1]] = conc[-1] # conc is the multiplier for inverse residence time to get the observed flux
xmod,cmod=invert( hmod,x0mod,pmod,ymod,rmod)

# now the preindustrial case
# here we only have the budget and reset the anthropogenic sources

x0pre = xmod.copy()
ppre = pmod.copy()


# zero anthropogenic sources
x0pre[3:6] = 0.
ppre[3:6] = 0.1 # and with small unc
hpre = np.ones(( 1, len( unknames)))
hpre[0, -3:] = -conc[1]
ypre = np.array([0.])
rpre = np.array( [10.])
xpre,cpre= invert( hpre, x0pre, ppre, ypre, rpre)
ind = unknames.index('kappacanopy')

x0lia = x0pre.copy()
plia = ppre.copy()
hlia = hpre.copy()
hlia[0, -3:] = -conc[0]
ylia = ypre.copy()
rlia = rpre.copy()
xlia,clia= invert( hlia, x0lia, plia, ylia, rlia)
print normal_ineq((xlia[ind], clia[ind,ind]**0.5), (xpre[ind], cpre[ind,ind]**0.5))
hdiff = hpre.copy()
x0diff = 0.*x0pre
pdiff = ppre.copy()
pdiff[0:8] *=0.3 # uncs of change are 10% of normal uncs
pdiff[8] = 0.1
pdiff[10] = 0.3
pdiff[9] = 0.3*xpre[9] # no limit on change in kappacanopy
pdiff[8:] = 0.2*xpre[8:]
# now the hardest part is the change in the budget engendered by changing ccn
ydiff = np.array( [xpre[-3:].sum() * (conc[0]- conc[1])])
rdiff = np.array( [abs(conc[0] -conc[1])*(cpre[-3:,-3:].sum())**0.5])
xdiff,cdiff = invert( hdiff, x0diff, pdiff, ydiff, rdiff)
xdiff2,cdiff2 = invert2( hdiff, x0diff, pdiff, ydiff, rdiff)
print (np.random.normal( xdiff[ind], cdiff[ind,ind]**0.5, 100000) > 0).sum()/100000.

